#pragma once
#include <luna/Action.h>
#include <luna/CircularQueue.h>
#include <luna/Spinlock.h>
#include <sys/types.h>

class Thread;

class Mutex
{
  public:
    void lock();
    void unlock();
    bool try_lock();

  private:
    CircularQueue<Thread*, 32> m_blocked_threads;
    Spinlock m_spinlock;
    Atomic<pid_t> m_thread;
};

template <typename T> class MutexLocked
{
    struct MutexLockedGuard
    {
        MutexLockedGuard(MutexLocked& value_ref) : m_value_ref(&value_ref)
        {
        }

        MutexLockedGuard(const MutexLockedGuard& other) = delete;
        MutexLockedGuard(MutexLockedGuard&& other)
        {
            m_value_ref = other.m_value_ref;
            other.m_value_ref = nullptr;
        }

        ~MutexLockedGuard()
        {
            if (m_value_ref) m_value_ref->m_lock.unlock();
        }

        T& ref()
        {
            expect(m_value_ref, "MutexLockedGuard::ref() called on a moved MutexLockedGuard");
            return m_value_ref->m_value;
        }

        void set(const T& other)
        {
            ref() = other;
        }

        T* operator->()
        {
            return &ref();
        }

        T& operator*()
        {
            return ref();
        }

      private:
        MutexLocked* m_value_ref;
    };

  public:
    MutexLocked() : m_value()
    {
    }

    MutexLocked(T value) : m_value(move(value))
    {
    }

    MutexLockedGuard lock()
    {
        m_lock.lock();
        return { *this };
    }

    Option<MutexLockedGuard> try_lock()
    {
        if (m_lock.try_lock()) { return { *this }; }
        return {};
    }

    void with_lock(Function<T&> callback)
    {
        m_lock.lock();
        callback(m_value);
        m_lock.unlock();
    }

  private:
    T m_value;
    Mutex m_lock;
};

class ScopedMutexLock
{
  public:
    ScopedMutexLock(Mutex& lock) : m_lock(lock)
    {
        m_lock.lock();
    }

    ~ScopedMutexLock()
    {
        if (!m_taken_over) m_lock.unlock();
    }

    ScopedMutexLock(const ScopedMutexLock&) = delete;
    ScopedMutexLock(ScopedMutexLock&&) = delete;

    Mutex& take_over()
    {
        m_taken_over = true;
        return m_lock;
    }

  private:
    Mutex& m_lock;
    bool m_taken_over { false };
};
