#include "Log.h"
#include "Pledge.h"
#include "memory/MemoryManager.h"
#include "sys/Syscall.h"
#include "thread/Scheduler.h"
#include <bits/waitpid.h>

Result<u64> sys_waitpid(Registers* regs, SyscallArgs args)
{
    pid_t pid = (pid_t)args[0];
    int* status_ptr = (int*)args[1];
    int options = (int)args[2];

    Thread* current = Scheduler::current();
    TRY(check_pledge(current->process, Promise::p_stdio));

    Process* target;

    if (pid > 0)
    {
        target = TRY(Result<Process*>::from_option(Scheduler::find_by_pid(pid), ESRCH));

        if (target->parent && target->parent != current->process) return err(ECHILD);

        if (options & WNOHANG) return err(EAGAIN);

    wait_for_child:
        if (!target->dead()) kernel_wait(pid);
        if (current->interrupted && (current->pending_signal_count() > 1 || current->pending_signal() != SIGCHLD))
        {
            kdbgln("signal: waitpid interrupted by signal");
            if (current->will_ignore_pending_signal())
            {
                current->process_pending_signals(regs);
                goto wait_for_child;
            }
            return err(EINTR);
        }

        check(target->dead());
    }
    else if (pid == -1)
    {
        if (!Scheduler::has_children(current->process)) return err(ECHILD);

        auto child = Scheduler::find_exited_child(current->process);
        if (!child.has_value())
        {
            if (options & WNOHANG) return err(EAGAIN);

        wait_for_any_child:
            kernel_wait(pid);
            if (current->interrupted && (current->pending_signal_count() > 1 || current->pending_signal() != SIGCHLD))
            {
                kdbgln("signal: waitpid interrupted by signal");
                if (current->will_ignore_pending_signal())
                {
                    current->process_pending_signals(regs);
                    goto wait_for_any_child;
                }
                return err(EINTR);
            }

            check(current->child_being_waited_for != -1);

            target = TRY(Result<Process*>::from_option(Scheduler::find_by_pid(current->child_being_waited_for), ESRCH));
            check(target->dead());
        }
        else
            target = child.value();
    }
    else // FIXME: Now that we have process groups, implement the cases where pid = 0 and pid < -1.
        return err(ENOTSUP);

    current->child_being_waited_for = -2;

    int status = (int)target->status;
    u64 id = target->id;

    current->process->user_ticks_children += target->user_ticks_self + target->user_ticks_children;
    current->process->kernel_ticks_children += target->kernel_ticks_self + target->kernel_ticks_children;

    target->thread_count = PROCESS_SHOULD_REAP;
    Scheduler::signal_reap_thread();

    if (status_ptr)
        if (!MemoryManager::copy_to_user_typed(status_ptr, &status)) return err(EFAULT);

    return id;
}
